from glue.core.data import BaseData
from glue.external.echo import CallbackProperty, SelectionCallbackProperty
from glue_vispy_viewers.common.viewer_state import Vispy3DViewerState
from glue.core.data_combo_helper import ManualDataComboHelper

__all__ = ['Vispy3DVolumeViewerState']


class Vispy3DVolumeViewerState(Vispy3DViewerState):

    downsample = CallbackProperty(True)
    resolution = SelectionCallbackProperty(4)
    reference_data = SelectionCallbackProperty(docstring='The dataset that is used to define the '
                                                         'available pixel/world components, and '
                                                         'which defines the coordinate frame in '
                                                         'which the images are shown')

    def __init__(self, **kwargs):

        super(Vispy3DVolumeViewerState, self).__init__()

        self.ref_data_helper = ManualDataComboHelper(self, 'reference_data')

        self.add_callback('layers', self._layers_changed)

        Vispy3DVolumeViewerState.resolution.set_choices(self, [2**i for i in range(4, 12)])

        self.update_from_dict(kwargs)

    def _first_3d_data(self):
        for layer_state in self.layers:
            if getattr(layer_state.layer, 'ndim', None) == 3:
                return layer_state.layer

    def _layers_changed(self, *args):
        self._update_combo_ref_data()
        self._set_reference_data()
        self._update_attributes()

    def _update_combo_ref_data(self, *args):
        self.ref_data_helper.set_multiple_data(self.layers_data)

    def _set_reference_data(self, *args):
        if self.reference_data is None:
            for layer in self.layers:
                if isinstance(layer.layer, BaseData):
                    self.reference_data = layer.layer
                    return

    def _update_attributes(self, *args):

        data = self._first_3d_data()

        if data is None:

            type(self).x_att.set_choices(self, [])
            type(self).y_att.set_choices(self, [])
            type(self).z_att.set_choices(self, [])

        else:

            z_cid, y_cid, x_cid = data.pixel_component_ids

            type(self).x_att.set_choices(self, [x_cid])
            type(self).y_att.set_choices(self, [y_cid])
            type(self).z_att.set_choices(self, [z_cid])

    @property
    def clip_limits_relative(self):

        data = self._first_3d_data()

        if data is None:
            return [0., 1., 0., 1., 0., 1.]
        else:
            nz, ny, nx = data.shape
            return (self.x_min / nx,
                    self.x_max / nx,
                    self.y_min / ny,
                    self.y_max / ny,
                    self.z_min / nz,
                    self.z_max / nz)



from __future__ import absolute_import, division, print_function

"""
This is for 3D selection in Glue 3d scatter plot viewer.
"""

import numpy as np

from glue.core import Data
from glue.config import viewer_tool

from glue.viewers.common.tool import CheckableTool

from glue.core.command import ApplySubsetState

from glue.core.roi import RectangularROI, CircularROI, PolygonalROI, Projected3dROI
from glue.core.subset import RoiSubsetState3d

from ..utils import as_matrix_transform
from ..extern.vispy.scene import Rectangle, Line, Ellipse

# Backward-compatibility for reading files
from .compat import MultiMaskSubsetState  # noqa
MultiElementSubsetState = MultiMaskSubsetState


class VispyMouseMode(CheckableTool):

    # this will create an abstract selection mode class to handle mouse events
    # instanced by lasso, rectangle, circular and point mode

    def __init__(self, viewer):
        super(VispyMouseMode, self).__init__(viewer)
        self._vispy_widget = viewer._vispy_widget
        self.current_visible_array = None

    def get_visible_data(self):
        visible = []
        # Loop over visible layer artists
        for layer_artist in self.viewer._layer_artist_container:
            # Only extract Data objects, not subsets
            if isinstance(layer_artist.layer, Data):
                visible.append(layer_artist.layer)
        visual = layer_artist.visual  # we only have one visual for each canvas
        return visible, visual

    def iter_data_layer_artists(self):
        for layer_artist in self.viewer._layer_artist_container:
            if isinstance(layer_artist.layer, Data):
                yield layer_artist

    def apply_roi(self, roi):
        x_att = self.viewer.state.x_att
        y_att = self.viewer.state.y_att
        z_att = self.viewer.state.z_att
        self.apply_subset_state(RoiSubsetState3d(x_att, y_att, z_att, roi))

    def apply_subset_state(self, subset_state):
        cmd = ApplySubsetState(data_collection=self.viewer._data,
                               subset_state=subset_state)
        self.viewer.session.command_stack.do(cmd)

    def set_progress(self, value):
        if value < 0:
            self.viewer.show_status('')
        else:
            self.viewer.show_status('Calculating selection - {0}%'.format(int(value)))

    @property
    def projection_matrix(self):

        # Get first layer (maybe just get from viewer directly in future)
        layer_artist = next(self.iter_data_layer_artists())

        # Get transformation matrix and transpose
        transform = layer_artist.visual.get_transform(map_from='visual', map_to='canvas')
        return as_matrix_transform(transform).matrix.T


@viewer_tool
class LassoSelectionMode(VispyMouseMode):

    icon = 'glue_lasso'
    tool_id = 'vispy:lasso'
    action_text = 'Select data using a lasso selection'

    def __init__(self, viewer):
        super(LassoSelectionMode, self).__init__(viewer)
        self.line = Line(color='purple',
                         width=2, method='agg',
                         parent=self._vispy_widget.canvas.scene)

    def activate(self):
        self.reset()

    def reset(self):
        self.line_pos = []
        self.line.set_data(np.zeros((0, 2), dtype=float))
        self.line.parent = None

    def press(self, event):
        if event.button == 1:
            self.line_pos.append(event.pos)

    def move(self, event):
        if event.button == 1 and event.is_dragging:
            self.line_pos.append(event.pos)
            self.line.set_data(np.array(self.line_pos, dtype=float))
            self.line.parent = self._vispy_widget.canvas.scene

    def release(self, event):

        if event.button == 1:

            if len(self.line_pos) > 0:
                vx, vy = np.array(self.line_pos).transpose()
                roi = Projected3dROI(roi_2d=PolygonalROI(vx, vy),
                                     projection_matrix=self.projection_matrix)
                self.apply_roi(roi)

            self.reset()

            self.viewer.toolbar.active_tool = None


@viewer_tool
class RectangleSelectionMode(VispyMouseMode):

    icon = 'glue_square'
    tool_id = 'vispy:rectangle'
    action_text = 'Select data using a rectangular selection'

    def __init__(self, viewer):
        super(RectangleSelectionMode, self).__init__(viewer)
        self.rectangle = Rectangle(center=(0, 0), width=1, height=1, border_width=2,
                                   color=(0, 0, 0, 0), border_color='purple')

    def activate(self):
        self.reset()

    def reset(self):
        self.corner1 = None
        self.corner2 = None
        self.rectangle.parent = None

    def press(self, event):
        if event.button == 1:
            self.corner1 = event.pos

    def move(self, event):
        if event.button == 1 and event.is_dragging:
            self.corner2 = event.pos
            x1, y1 = self.corner1
            x2, y2 = self.corner2
            if abs(x2 - x1) > 0 and abs(y2 - y1) > 0:
                self.rectangle.center = 0.5 * (x1 + x2), 0.5 * (y1 + y2)
                self.rectangle.width = abs(x2 - x1)
                self.rectangle.height = abs(y2 - y1)
                self.rectangle.parent = self._vispy_widget.canvas.scene

    @property
    def bounds(self):
        x1, y1 = self.corner1
        x2, y2 = self.corner2
        return (min(x1, x2), max(x1, x2), min(y1, y2), max(y1, y2))

    def release(self, event):

        if event.button == 1:

            if self.corner2 is not None:
                roi = Projected3dROI(roi_2d=RectangularROI(*self.bounds),
                                     projection_matrix=self.projection_matrix)
                self.apply_roi(roi)

            self.reset()

            self.viewer.toolbar.active_tool = None


@viewer_tool
class CircleSelectionMode(VispyMouseMode):

    icon = 'glue_circle'
    tool_id = 'vispy:circle'
    action_text = 'Select data using a circular selection'

    def __init__(self, viewer):
        super(CircleSelectionMode, self).__init__(viewer)
        self.ellipse = Ellipse(center=(0, 0), radius=1, border_width=2,
                               color=(0, 0, 0, 0), border_color='purple')

    def activate(self):
        self.reset()

    def reset(self):
        self.center = None
        self.radius = 0
        self.ellipse.parent = None

    def press(self, event):
        if event.button == 1:
            self.center = event.pos

    def move(self, event):
        if event.button == 1 and event.is_dragging:
            self.radius = max(abs(event.pos[0] - self.center[0]),
                              abs(event.pos[1] - self.center[1]))
            if self.radius > 0:
                self.ellipse.center = self.center
                self.ellipse.radius = self.radius
                self.ellipse.parent = self._vispy_widget.canvas.scene

    def release(self, event):

        if event.button == 1:

            if self.radius > 0:
                roi = Projected3dROI(roi_2d=CircularROI(self.center[0],
                                                        self.center[1],
                                                        self.radius),
                                     projection_matrix=self.projection_matrix)
                self.apply_roi(roi)

            self.reset()

            self.viewer.toolbar.active_tool = None



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014, Nicolas P. Rougier
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------
"""
A collection is a container for several (optionally indexed) objects having
the same vertex structure (vtype) and same uniforms type (utype). A collection
allows to manipulate objects individually and each object can have its own set
of uniforms provided they are a combination of floats.
"""

from __future__ import division

import math
import numpy as np
from ...gloo import Texture2D, VertexBuffer, IndexBuffer
from . util import dtype_reduce
from . array_list import ArrayList


def next_power_of_2(n):
    """ Return next power of 2 greater than or equal to n """
    n -= 1  # greater than OR EQUAL TO n
    shift = 1
    while (n + 1) & n:  # n+1 is not a power of 2 yet
        n |= n >> shift
        shift *= 2
    return max(4, n + 1)


class Item(object):

    """
    An item represent an object within a collection and is created on demand
    when accessing a specific object of the collection.
    """

    def __init__(self, parent, key, vertices, indices, uniforms):
        """
        Create an item from an existing collection.

        Parameters
        ----------

        parent : Collection
            Collection this item belongs to

        key : int
            Collection index of this item

        vertices: array-like
            Vertices of the item

        indices: array-like
            Indices of the item

        uniforms: array-like
            Uniform parameters of the item
        """

        self._parent = parent
        self._key = key
        self._vertices = vertices
        self._indices = indices
        self._uniforms = uniforms

    @property
    def vertices(self):
        return self._vertices

    @vertices.setter
    def vertices(self, data):
        self._vertices[...] = np.array(data)

    @property
    def indices(self):
        return self._indices

    @indices.setter
    def indices(self, data):
        if self._indices is None:
            raise ValueError("Item has no indices")
        start = self._parent.vertices._items[self._key][0]
        self._indices[...] = np.array(data) + start

    @property
    def uniforms(self):
        return self._uniforms

    @uniforms.setter
    def uniforms(self, data):
        if self._uniforms is None:
            raise ValueError("Item has no associated uniform")
        self._uniforms[...] = data

    def __getitem__(self, key):
        """ Get a specific uniforms value """

        if key in self._vertices.dtype.names:
            return self._vertices[key]
        elif key in self._uniforms.dtype.names:
            return self._uniforms[key]
        else:
            raise IndexError("Unknown key ('%s')" % key)

    def __setitem__(self, key, value):
        """ Set a specific uniforms value """

        if key in self._vertices.dtype.names:
            self._vertices[key] = value
        elif key in self._uniforms.dtype.names:
            self._uniforms[key] = value
        else:
            raise IndexError("Unknown key")

    def __str__(self):
        return "Item (%s, %s, %s)" % (self._vertices,
                                      self._indices,
                                      self._uniforms)


class BaseCollection(object):

    def __init__(self, vtype, utype=None, itype=None):

        # Vertices and type (mandatory)
        self._vertices_list = None
        self._vertices_buffer = None

        # Vertex indices and type (optional)
        self._indices_list = None
        self._indices_buffer = None

        # Uniforms and type (optional)
        self._uniforms_list = None
        self._uniforms_texture = None

        # Make sure types are np.dtype (or None)
        vtype = np.dtype(vtype) if vtype is not None else None
        itype = np.dtype(itype) if itype is not None else None
        utype = np.dtype(utype) if utype is not None else None

        # Vertices type (mandatory)
        # -------------------------
        if vtype.names is None:
            raise ValueError("vtype must be a structured dtype")

        # Indices type (optional)
        # -----------------------
        if itype is not None:
            if itype not in [np.uint8, np.uint16, np.uint32]:
                raise ValueError("itype must be unsigned integer or None")
            self._indices_list = ArrayList(dtype=itype)

        # No program yet
        self._programs = []

        # Need to update buffers & texture
        self._need_update = True

        # Uniforms type (optional)
        # -------------------------
        if utype is not None:

            if utype.names is None:
                raise ValueError("utype must be a structured dtype")

            # Convert types to lists (in case they were already dtypes) such
            # that we can append new fields
            vtype = eval(str(np.dtype(vtype)))
            # We add a uniform index to access uniform data
            vtype.append(('collection_index', np.float32))
            vtype = np.dtype(vtype)

            # Check utype is made of float32 only
            utype = eval(str(np.dtype(utype)))
            r_utype = dtype_reduce(utype)
            if type(r_utype[0]) is not str or r_utype[2] != 'float32':
                raise RuntimeError("utype cannot be reduced to float32 only")

            # Make utype divisible by 4
            # count = ((r_utype[1]-1)//4+1)*4

            # Make utype a power of two
            count = next_power_of_2(r_utype[1])
            if (count - r_utype[1]) > 0:
                utype.append(('__unused__', np.float32, count - r_utype[1]))

            self._uniforms_list = ArrayList(dtype=utype)
            self._uniforms_float_count = count

            # Reserve some space in texture such that we have
            # at least one full line
            shape = self._compute_texture_shape(1)
            self._uniforms_list.reserve(shape[1] / (count / 4))

        # Last since utype may add a new field in vtype (collecion_index)
        self._vertices_list = ArrayList(dtype=vtype)

        # Record all types
        self._vtype = np.dtype(vtype)
        self._itype = np.dtype(itype) if itype is not None else None
        self._utype = np.dtype(utype) if utype is not None else None

    def __len__(self):
        """ x.__len__() <==> len(x) """

        return len(self._vertices_list)

    @property
    def vtype(self):
        """ Vertices dtype """

        return self._vtype

    @property
    def itype(self):
        """ Indices dtype """

        return self._itype

    @property
    def utype(self):
        """ Uniforms dtype """

        return self._utype

    def append(self, vertices, uniforms=None, indices=None, itemsize=None):
        """
        Parameters
        ----------

        vertices : numpy array
            An array whose dtype is compatible with self.vdtype

        uniforms: numpy array
            An array whose dtype is compatible with self.utype

        indices : numpy array
            An array whose dtype is compatible with self.idtype
            All index values must be between 0 and len(vertices)

        itemsize: int, tuple or 1-D array
            If `itemsize is an integer, N, the array will be divided
            into elements of size N. If such partition is not possible,
            an error is raised.

            If `itemsize` is 1-D array, the array will be divided into
            elements whose succesive sizes will be picked from itemsize.
            If the sum of itemsize values is different from array size,
            an error is raised.
        """

        # Vertices
        # -----------------------------
        vertices = np.array(vertices).astype(self.vtype).ravel()
        vsize = self._vertices_list.size

        # No itemsize given
        # -----------------
        if itemsize is None:
            index = 0
            count = 1

        # Uniform itemsize (int)
        # ----------------------
        elif isinstance(itemsize, int):
            count = len(vertices) / itemsize
            index = np.repeat(np.arange(count), itemsize)

        # Individual itemsize (array)
        # ---------------------------
        elif isinstance(itemsize, (np.ndarray, list)):
            count = len(itemsize)
            index = np.repeat(np.arange(count), itemsize)
        else:
            raise ValueError("Itemsize not understood")

        if self.utype:
            vertices["collection_index"] = index + len(self)
        self._vertices_list.append(vertices, itemsize)

        # Indices
        # -----------------------------
        if self.itype is not None:
            # No indices given (-> automatic generation)
            if indices is None:
                indices = vsize + np.arange(len(vertices))
                self._indices_list.append(indices, itemsize)

            # Indices given
            # FIXME: variables indices (list of list or ArrayList)
            else:
                if itemsize is None:
                    I = np.array(indices) + vsize
                elif isinstance(itemsize, int):
                    I = vsize + (np.tile(indices, count) +
                                 itemsize * np.repeat(np.arange(count), len(indices)))  # noqa
                else:
                    raise ValueError("Indices not compatible with items")
                self._indices_list.append(I, len(indices))

        # Uniforms
        # -----------------------------
        if self.utype:
            if uniforms is None:
                uniforms = np.zeros(count, dtype=self.utype)
            else:
                uniforms = np.array(uniforms).astype(self.utype).ravel()
            self._uniforms_list.append(uniforms, itemsize=1)

        self._need_update = True

    def __delitem__(self, index):
        """ x.__delitem__(y) <==> del x[y] """

        # Deleting one item
        if isinstance(index, int):
            if index < 0:
                index += len(self)
            if index < 0 or index > len(self):
                raise IndexError("Collection deletion index out of range")
            istart, istop = index, index + 1
        # Deleting several items
        elif isinstance(index, slice):
            istart, istop, _ = index.indices(len(self))
            if istart > istop:
                istart, istop = istop, istart
            if istart == istop:
                return
        # Deleting everything
        elif index is Ellipsis:
            istart, istop = 0, len(self)
        # Error
        else:
            raise TypeError("Collection deletion indices must be integers")

        vsize = len(self._vertices_list[index])
        if self.itype is not None:
            del self._indices_list[index]
            self._indices_list[index:] -= vsize

        if self.utype:
            self._vertices_list[index:]["collection_index"] -= istop - istart
        del self._vertices_list[index]

        if self.utype is not None:
            del self._uniforms_list[index]

        self._need_update = True

    def __getitem__(self, key):
        """ """

        # WARNING
        # Here we want to make sure to use buffers and texture (instead of
        # lists) since only them are aware of any external modification.
        if self._need_update:
            self._update()

        V = self._vertices_buffer
        I = None
        U = None
        if self._indices_list is not None:
            I = self._indices_buffer
        if self._uniforms_list is not None:
            U = self._uniforms_texture.data.ravel().view(self.utype)

        # Getting a whole field
        if isinstance(key, str):
            # Getting a named field from vertices
            if key in V.dtype.names:
                return V[key]
            # Getting a named field from uniforms
            elif U is not None and key in U.dtype.names:
                # Careful, U is the whole texture that can be bigger than list
                # return U[key]
                return U[key][:len(self._uniforms_list)]
            else:
                raise IndexError("Unknown field name ('%s')" % key)

        # Getting individual item
        elif isinstance(key, int):
            vstart, vend = self._vertices_list._items[key]
            vertices = V[vstart:vend]
            indices = None
            uniforms = None
            if I is not None:
                istart, iend = self._indices_list._items[key]
                indices = I[istart:iend]

            if U is not None:
                ustart, uend = self._uniforms_list._items[key]
                uniforms = U[ustart:uend]

            return Item(self, key, vertices, indices, uniforms)

        # Error
        else:
            raise IndexError("Cannot get more than one item at once")

    def __setitem__(self, key, data):
        """ x.__setitem__(i, y) <==> x[i]=y """

        # if len(self._programs):
        # found = False
        # for program in self._programs:
        #     if key in program.hooks:
        #         program[key] = data
        #         found = True
        # if found: return

        # WARNING
        # Here we want to make sure to use buffers and texture (instead of
        # lists) since only them are aware of any external modification.
        if self._need_update:
            self._update()

        V = self._vertices_buffer
        I = None
        U = None
        if self._indices_list is not None:
            I = self._indices_buffer  # noqa
        if self._uniforms_list is not None:
            U = self._uniforms_texture.data.ravel().view(self.utype)

        # Setting a whole field
        if isinstance(key, str):
            # Setting a named field in vertices
            if key in self.vtype.names:
                V[key] = data
            # Setting a named field in uniforms
            elif self.utype and key in self.utype.names:
                # Careful, U is the whole texture that can be bigger than list
                # U[key] = data
                U[key][:len(self._uniforms_list)] = data
            else:
                raise IndexError("Unknown field name ('%s')" % key)

        # # Setting individual item
        # elif isinstance(key, int):
        #     #vstart, vend = self._vertices_list._items[key]
        #     #istart, iend = self._indices_list._items[key]
        #     #ustart, uend = self._uniforms_list._items[key]
        #     vertices, indices, uniforms = data
        #     del self[key]
        #     self.insert(key, vertices, indices, uniforms)

        else:
            raise IndexError("Cannot set more than one item")

    def _compute_texture_shape(self, size=1):
        """ Compute uniform texture shape """

        # We should use this line but we may not have a GL context yet
        # linesize = gl.glGetInteger(gl.GL_MAX_TEXTURE_SIZE)
        linesize = 1024
        count = self._uniforms_float_count
        cols = 4 * linesize // int(count)
        rows = max(1, int(math.ceil(size / float(cols))))
        shape = rows, cols * (count // 4), count
        self._ushape = shape
        return shape

    def _update(self):
        """ Update vertex buffers & texture """

        if self._vertices_buffer is not None:
            self._vertices_buffer.delete()
        self._vertices_buffer = VertexBuffer(self._vertices_list.data)

        if self.itype is not None:
            if self._indices_buffer is not None:
                self._indices_buffer.delete()
            self._indices_buffer = IndexBuffer(self._indices_list.data)

        if self.utype is not None:
            if self._uniforms_texture is not None:
                self._uniforms_texture.delete()

            # We take the whole array (_data), not the data one
            texture = self._uniforms_list._data.view(np.float32)
            size = len(texture) / self._uniforms_float_count
            shape = self._compute_texture_shape(size)

            # shape[2] = float count is only used in vertex shader code
            texture = texture.reshape(shape[0], shape[1], 4)
            self._uniforms_texture = Texture2D(texture)
            self._uniforms_texture.data = texture
            self._uniforms_texture.interpolation = 'nearest'

        if len(self._programs):
            for program in self._programs:
                program.bind(self._vertices_buffer)
                if self._uniforms_list is not None:
                    program["uniforms"] = self._uniforms_texture
                    program["uniforms_shape"] = self._ushape



"""

THIS CODE IS AUTO-GENERATED. DO NOT EDIT.

Base proxy API for GL ES 2.0.

"""

class BaseGLProxy(object):
    """ Base proxy class for the GL ES 2.0 API. Subclasses should
    implement __call__ to process the API calls.
    """
   
    def __call__(self, funcname, returns, *args):
        raise NotImplementedError()


    def glActiveTexture(self, texture):
        self("glActiveTexture", False, texture)


    def glAttachShader(self, program, shader):
        self("glAttachShader", False, program, shader)


    def glBindAttribLocation(self, program, index, name):
        self("glBindAttribLocation", False, program, index, name)


    def glBindBuffer(self, target, buffer):
        self("glBindBuffer", False, target, buffer)


    def glBindFramebuffer(self, target, framebuffer):
        self("glBindFramebuffer", False, target, framebuffer)


    def glBindRenderbuffer(self, target, renderbuffer):
        self("glBindRenderbuffer", False, target, renderbuffer)


    def glBindTexture(self, target, texture):
        self("glBindTexture", False, target, texture)


    def glBlendColor(self, red, green, blue, alpha):
        self("glBlendColor", False, red, green, blue, alpha)


    def glBlendEquation(self, mode):
        self("glBlendEquation", False, mode)


    def glBlendEquationSeparate(self, modeRGB, modeAlpha):
        self("glBlendEquationSeparate", False, modeRGB, modeAlpha)


    def glBlendFunc(self, sfactor, dfactor):
        self("glBlendFunc", False, sfactor, dfactor)


    def glBlendFuncSeparate(self, srcRGB, dstRGB, srcAlpha, dstAlpha):
        self("glBlendFuncSeparate", False, srcRGB, dstRGB, srcAlpha, dstAlpha)


    def glBufferData(self, target, data, usage):
        self("glBufferData", False, target, data, usage)


    def glBufferSubData(self, target, offset, data):
        self("glBufferSubData", False, target, offset, data)


    def glCheckFramebufferStatus(self, target):
        return self("glCheckFramebufferStatus", True, target)


    def glClear(self, mask):
        self("glClear", False, mask)


    def glClearColor(self, red, green, blue, alpha):
        self("glClearColor", False, red, green, blue, alpha)


    def glClearDepth(self, depth):
        self("glClearDepth", False, depth)


    def glClearStencil(self, s):
        self("glClearStencil", False, s)


    def glColorMask(self, red, green, blue, alpha):
        self("glColorMask", False, red, green, blue, alpha)


    def glCompileShader(self, shader):
        self("glCompileShader", False, shader)


    def glCompressedTexImage2D(self, target, level, internalformat, width, height, border, data):
        self("glCompressedTexImage2D", False, target, level, internalformat, width, height, border, data)


    def glCompressedTexSubImage2D(self, target, level, xoffset, yoffset, width, height, format, data):
        self("glCompressedTexSubImage2D", False, target, level, xoffset, yoffset, width, height, format, data)


    def glCopyTexImage2D(self, target, level, internalformat, x, y, width, height, border):
        self("glCopyTexImage2D", False, target, level, internalformat, x, y, width, height, border)


    def glCopyTexSubImage2D(self, target, level, xoffset, yoffset, x, y, width, height):
        self("glCopyTexSubImage2D", False, target, level, xoffset, yoffset, x, y, width, height)


    def glCreateProgram(self, ):
        return self("glCreateProgram", True, )


    def glCreateShader(self, type):
        return self("glCreateShader", True, type)


    def glCullFace(self, mode):
        self("glCullFace", False, mode)


    def glDeleteBuffer(self, buffer):
        self("glDeleteBuffer", False, buffer)


    def glDeleteFramebuffer(self, framebuffer):
        self("glDeleteFramebuffer", False, framebuffer)


    def glDeleteProgram(self, program):
        self("glDeleteProgram", False, program)


    def glDeleteRenderbuffer(self, renderbuffer):
        self("glDeleteRenderbuffer", False, renderbuffer)


    def glDeleteShader(self, shader):
        self("glDeleteShader", False, shader)


    def glDeleteTexture(self, texture):
        self("glDeleteTexture", False, texture)


    def glDepthFunc(self, func):
        self("glDepthFunc", False, func)


    def glDepthMask(self, flag):
        self("glDepthMask", False, flag)


    def glDepthRange(self, zNear, zFar):
        self("glDepthRange", False, zNear, zFar)


    def glDetachShader(self, program, shader):
        self("glDetachShader", False, program, shader)


    def glDisable(self, cap):
        self("glDisable", False, cap)


    def glDisableVertexAttribArray(self, index):
        self("glDisableVertexAttribArray", False, index)


    def glDrawArrays(self, mode, first, count):
        self("glDrawArrays", False, mode, first, count)


    def glDrawElements(self, mode, count, type, offset):
        self("glDrawElements", False, mode, count, type, offset)


    def glEnable(self, cap):
        self("glEnable", False, cap)


    def glEnableVertexAttribArray(self, index):
        self("glEnableVertexAttribArray", False, index)


    def glFinish(self, ):
        self("glFinish", False, )


    def glFlush(self, ):
        self("glFlush", False, )


    def glFramebufferRenderbuffer(self, target, attachment, renderbuffertarget, renderbuffer):
        self("glFramebufferRenderbuffer", False, target, attachment, renderbuffertarget, renderbuffer)


    def glFramebufferTexture2D(self, target, attachment, textarget, texture, level):
        self("glFramebufferTexture2D", False, target, attachment, textarget, texture, level)


    def glFrontFace(self, mode):
        self("glFrontFace", False, mode)


    def glCreateBuffer(self, ):
        return self("glCreateBuffer", True, )


    def glCreateFramebuffer(self, ):
        return self("glCreateFramebuffer", True, )


    def glCreateRenderbuffer(self, ):
        return self("glCreateRenderbuffer", True, )


    def glCreateTexture(self, ):
        return self("glCreateTexture", True, )


    def glGenerateMipmap(self, target):
        self("glGenerateMipmap", False, target)


    def glGetActiveAttrib(self, program, index):
        return self("glGetActiveAttrib", True, program, index)


    def glGetActiveUniform(self, program, index):
        return self("glGetActiveUniform", True, program, index)


    def glGetAttachedShaders(self, program):
        return self("glGetAttachedShaders", True, program)


    def glGetAttribLocation(self, program, name):
        return self("glGetAttribLocation", True, program, name)


    def _glGetBooleanv(self, pname):
        self("_glGetBooleanv", False, pname)


    def glGetBufferParameter(self, target, pname):
        return self("glGetBufferParameter", True, target, pname)


    def glGetError(self, ):
        return self("glGetError", True, )


    def _glGetFloatv(self, pname):
        self("_glGetFloatv", False, pname)


    def glGetFramebufferAttachmentParameter(self, target, attachment, pname):
        return self("glGetFramebufferAttachmentParameter", True, target, attachment, pname)


    def _glGetIntegerv(self, pname):
        self("_glGetIntegerv", False, pname)


    def glGetProgramInfoLog(self, program):
        return self("glGetProgramInfoLog", True, program)


    def glGetProgramParameter(self, program, pname):
        return self("glGetProgramParameter", True, program, pname)


    def glGetRenderbufferParameter(self, target, pname):
        return self("glGetRenderbufferParameter", True, target, pname)


    def glGetShaderInfoLog(self, shader):
        return self("glGetShaderInfoLog", True, shader)


    def glGetShaderPrecisionFormat(self, shadertype, precisiontype):
        return self("glGetShaderPrecisionFormat", True, shadertype, precisiontype)


    def glGetShaderSource(self, shader):
        return self("glGetShaderSource", True, shader)


    def glGetShaderParameter(self, shader, pname):
        return self("glGetShaderParameter", True, shader, pname)


    def glGetParameter(self, pname):
        return self("glGetParameter", True, pname)


    def glGetTexParameter(self, target, pname):
        return self("glGetTexParameter", True, target, pname)


    def glGetUniform(self, program, location):
        return self("glGetUniform", True, program, location)


    def glGetUniformLocation(self, program, name):
        return self("glGetUniformLocation", True, program, name)


    def glGetVertexAttrib(self, index, pname):
        return self("glGetVertexAttrib", True, index, pname)


    def glGetVertexAttribOffset(self, index, pname):
        return self("glGetVertexAttribOffset", True, index, pname)


    def glHint(self, target, mode):
        self("glHint", False, target, mode)


    def glIsBuffer(self, buffer):
        return self("glIsBuffer", True, buffer)


    def glIsEnabled(self, cap):
        return self("glIsEnabled", True, cap)


    def glIsFramebuffer(self, framebuffer):
        return self("glIsFramebuffer", True, framebuffer)


    def glIsProgram(self, program):
        return self("glIsProgram", True, program)


    def glIsRenderbuffer(self, renderbuffer):
        return self("glIsRenderbuffer", True, renderbuffer)


    def glIsShader(self, shader):
        return self("glIsShader", True, shader)


    def glIsTexture(self, texture):
        return self("glIsTexture", True, texture)


    def glLineWidth(self, width):
        self("glLineWidth", False, width)


    def glLinkProgram(self, program):
        self("glLinkProgram", False, program)


    def glPixelStorei(self, pname, param):
        self("glPixelStorei", False, pname, param)


    def glPolygonOffset(self, factor, units):
        self("glPolygonOffset", False, factor, units)


    def glReadPixels(self, x, y, width, height, format, type):
        return self("glReadPixels", True, x, y, width, height, format, type)


    def glRenderbufferStorage(self, target, internalformat, width, height):
        self("glRenderbufferStorage", False, target, internalformat, width, height)


    def glSampleCoverage(self, value, invert):
        self("glSampleCoverage", False, value, invert)


    def glScissor(self, x, y, width, height):
        self("glScissor", False, x, y, width, height)


    def glShaderSource(self, shader, source):
        self("glShaderSource", False, shader, source)


    def glStencilFunc(self, func, ref, mask):
        self("glStencilFunc", False, func, ref, mask)


    def glStencilFuncSeparate(self, face, func, ref, mask):
        self("glStencilFuncSeparate", False, face, func, ref, mask)


    def glStencilMask(self, mask):
        self("glStencilMask", False, mask)


    def glStencilMaskSeparate(self, face, mask):
        self("glStencilMaskSeparate", False, face, mask)


    def glStencilOp(self, fail, zfail, zpass):
        self("glStencilOp", False, fail, zfail, zpass)


    def glStencilOpSeparate(self, face, fail, zfail, zpass):
        self("glStencilOpSeparate", False, face, fail, zfail, zpass)


    def glTexImage2D(self, target, level, internalformat, format, type, pixels):
        self("glTexImage2D", False, target, level, internalformat, format, type, pixels)


    def glTexParameterf(self, target, pname, param):
        self("glTexParameterf", False, target, pname, param)
    def glTexParameteri(self, target, pname, param):
        self("glTexParameteri", False, target, pname, param)


    def glTexSubImage2D(self, target, level, xoffset, yoffset, format, type, pixels):
        self("glTexSubImage2D", False, target, level, xoffset, yoffset, format, type, pixels)


    def glUniform1f(self, location, v1):
        self("glUniform1f", False, location, v1)
    def glUniform2f(self, location, v1, v2):
        self("glUniform2f", False, location, v1, v2)
    def glUniform3f(self, location, v1, v2, v3):
        self("glUniform3f", False, location, v1, v2, v3)
    def glUniform4f(self, location, v1, v2, v3, v4):
        self("glUniform4f", False, location, v1, v2, v3, v4)
    def glUniform1i(self, location, v1):
        self("glUniform1i", False, location, v1)
    def glUniform2i(self, location, v1, v2):
        self("glUniform2i", False, location, v1, v2)
    def glUniform3i(self, location, v1, v2, v3):
        self("glUniform3i", False, location, v1, v2, v3)
    def glUniform4i(self, location, v1, v2, v3, v4):
        self("glUniform4i", False, location, v1, v2, v3, v4)
    def glUniform1fv(self, location, count, values):
        self("glUniform1fv", False, location, count, values)
    def glUniform2fv(self, location, count, values):
        self("glUniform2fv", False, location, count, values)
    def glUniform3fv(self, location, count, values):
        self("glUniform3fv", False, location, count, values)
    def glUniform4fv(self, location, count, values):
        self("glUniform4fv", False, location, count, values)
    def glUniform1iv(self, location, count, values):
        self("glUniform1iv", False, location, count, values)
    def glUniform2iv(self, location, count, values):
        self("glUniform2iv", False, location, count, values)
    def glUniform3iv(self, location, count, values):
        self("glUniform3iv", False, location, count, values)
    def glUniform4iv(self, location, count, values):
        self("glUniform4iv", False, location, count, values)


    def glUniformMatrix2fv(self, location, count, transpose, values):
        self("glUniformMatrix2fv", False, location, count, transpose, values)
    def glUniformMatrix3fv(self, location, count, transpose, values):
        self("glUniformMatrix3fv", False, location, count, transpose, values)
    def glUniformMatrix4fv(self, location, count, transpose, values):
        self("glUniformMatrix4fv", False, location, count, transpose, values)


    def glUseProgram(self, program):
        self("glUseProgram", False, program)


    def glValidateProgram(self, program):
        self("glValidateProgram", False, program)


    def glVertexAttrib1f(self, index, v1):
        self("glVertexAttrib1f", False, index, v1)
    def glVertexAttrib2f(self, index, v1, v2):
        self("glVertexAttrib2f", False, index, v1, v2)
    def glVertexAttrib3f(self, index, v1, v2, v3):
        self("glVertexAttrib3f", False, index, v1, v2, v3)
    def glVertexAttrib4f(self, index, v1, v2, v3, v4):
        self("glVertexAttrib4f", False, index, v1, v2, v3, v4)


    def glVertexAttribPointer(self, indx, size, type, normalized, stride, offset):
        self("glVertexAttribPointer", False, indx, size, type, normalized, stride, offset)


    def glViewport(self, x, y, width, height):
        self("glViewport", False, x, y, width, height)





import numpy as np

from glue.core import DataCollection, Data
from glue.app.qt.application import GlueApplication
from glue.core.component import Component
from glue.core.link_helpers import LinkSame
from glue.core.fixed_resolution_buffer import PIXEL_CACHE, ARRAY_CACHE

from ..volume_viewer import VispyVolumeViewer


def teardown_function(function):
    # Make sure cache is empty
    if len(PIXEL_CACHE) > 0:
        raise Exception("Pixel cache contains {0} elements".format(len(PIXEL_CACHE)))
    if len(ARRAY_CACHE) > 0:
        raise Exception("Array cache contains {0} elements".format(len(ARRAY_CACHE)))


def make_test_data(dimensions=(10, 10, 10)):

    data = Data(label="Test Cube Data")

    np.random.seed(12345)

    for letter in 'abc':
        comp = Component(np.random.random(dimensions))
        data.add_component(comp, letter)

    return data


def test_volume_viewer(tmpdir):

    # Create fake data
    data = make_test_data()

    # Create fake session

    dc = DataCollection([data])
    ga = GlueApplication(dc)
    ga.show()

    volume = ga.new_data_viewer(VispyVolumeViewer)
    volume.add_data(data)
    volume.viewer_size = (400, 500)

    viewer_state = volume.state

    viewer_state.x_stretch = 0.5
    viewer_state.y_stretch = 1.0
    viewer_state.z_stretch = 2.0

    viewer_state.x_min = -0.1
    viewer_state.x_max = 10.1
    viewer_state.y_min = 0.1
    viewer_state.y_max = 10.9
    viewer_state.z_min = 0.2
    viewer_state.z_max = 10.8

    viewer_state.visible_axes = False

    # Get layer artist style editor
    layer_state = viewer_state.layers[0]

    layer_state.attribute = data.id['b']
    layer_state.vmin = 0.1
    layer_state.vmax = 0.9
    layer_state.alpha = 0.8

    # Check that writing a session works as expected.

    session_file = tmpdir.join('test_volume_viewer.glu').strpath
    ga.save_session(session_file)
    ga.close()

    # Now we can check that everything is restored correctly

    ga2 = GlueApplication.restore_session(session_file)
    ga2.show()

    volume_r = ga2.viewers[0][0]

    assert volume_r.viewer_size == (400, 500)

    viewer_state = volume_r.state

    np.testing.assert_allclose(viewer_state.x_stretch, 0.5, rtol=1e-3)
    np.testing.assert_allclose(viewer_state.y_stretch, 1.0, rtol=1e-3)
    np.testing.assert_allclose(viewer_state.z_stretch, 2.0, rtol=1e-3)

    assert viewer_state.x_min == -0.1
    assert viewer_state.x_max == 10.1
    assert viewer_state.y_min == 0.1
    assert viewer_state.y_max == 10.9
    assert viewer_state.z_min == 0.2
    assert viewer_state.z_max == 10.8

    assert not viewer_state.visible_axes

    layer_artist = viewer_state.layers[0]

    assert layer_artist.attribute.label == 'b'
    assert layer_artist.vmin == 0.1
    assert layer_artist.vmax == 0.9
    assert layer_artist.alpha == 0.8

    ga2.close()


def test_array_shape(tmpdir):
    # Create irregularly shaped data cube
    data = make_test_data((3841, 48, 46))

    # Create fake session

    dc = DataCollection([data])
    ga = GlueApplication(dc)

    volume = ga.new_data_viewer(VispyVolumeViewer)
    volume.add_data(data)

    viewer_state = volume.state

    # Get layer artist style editor
    layer_state = viewer_state.layers[0]

    layer_state.attribute = data.id['b']

    ga.close()


def test_scatter_on_volume(tmpdir):

    data1 = Data(a=np.arange(60).reshape((3, 4, 5)))
    data2 = Data(x=[1, 2, 3], y=[2, 3, 4], z=[3, 4, 5])
    data3 = Data(b=np.arange(60).reshape((3, 4, 5)))

    dc = DataCollection([data1, data2, data3])

    dc.add_link(LinkSame(data1.pixel_component_ids[2], data2.id['x']))
    dc.add_link(LinkSame(data1.pixel_component_ids[1], data2.id['y']))
    dc.add_link(LinkSame(data1.pixel_component_ids[0], data2.id['z']))

    ga = GlueApplication(dc)
    ga.show()

    volume = ga.new_data_viewer(VispyVolumeViewer)
    volume.add_data(data1)
    volume.add_data(data2)
    volume.add_data(data3)

    # Check that writing a session works as expected.

    session_file = tmpdir.join('test_scatter_on_volume.glu').strpath
    ga.save_session(session_file)
    ga.close()

    # Now we can check that everything is restored correctly

    ga2 = GlueApplication.restore_session(session_file)
    ga2.show()

    volume_r = ga2.viewers[0][0]

    assert len(volume_r.layers) == 3

    ga2.close()


def test_layer_visibility_clip():

    # Regression test for a bug that meant that updating the clip data setting
    # caused a layer to become visible even if it shouldn't be

    # Create fake data
    data = make_test_data()

    # Create fake session

    dc = DataCollection([data])
    ga = GlueApplication(dc)
    ga.show()

    volume = ga.new_data_viewer(VispyVolumeViewer)
    volume.add_data(data)

    assert volume.layers[0].visible
    assert volume.layers[0]._multivol.enabled[0]

    volume.layers[0].visible = False

    assert not volume.layers[0].visible
    assert not volume.layers[0]._multivol.enabled[0]

    volume.state.clip_data = True

    assert not volume.layers[0].visible
    assert not volume.layers[0]._multivol.enabled[0]

    ga.close()


def test_remove_subset_group():

    # Regression test for a bug that meant that removing a subset caused an
    # error when multiple viewers were present.

    # Create fake data
    data = make_test_data()

    # Create fake session

    dc = DataCollection([data])
    ga = GlueApplication(dc)
    ga.show()

    volume1 = ga.new_data_viewer(VispyVolumeViewer)
    volume1.add_data(data)

    volume2 = ga.new_data_viewer(VispyVolumeViewer)
    volume2.add_data(data)

    dc.new_subset_group(subset_state=data.id['a'] > 0, label='Subset 1')
    dc.remove_subset_group(dc.subset_groups[0])

    ga.close()


def test_add_data_with_incompatible_subsets(tmpdir):

    data1 = Data(label="Data 1", x=np.arange(24).reshape((4, 3, 2)))
    data2 = Data(label="Data 2", y=np.arange(24).reshape((4, 3, 2)))

    dc = DataCollection([data1, data2])
    ga = GlueApplication(dc)
    ga.show()

    # Subset is defined in terms of data2, so it's an incompatible subset
    # for data1
    dc.new_subset_group(subset_state=data2.id['y'] > 0.5, label='subset 1')

    volume = ga.new_data_viewer(VispyVolumeViewer)
    volume.add_data(data1)

    ga.close()



from __future__ import absolute_import, division, print_function

from glue.external.echo import CallbackProperty, keep_in_sync
from glue.core.message import LayerArtistUpdatedMessage
from glue.viewers.common.state import LayerState

__all__ = ['VispyLayerState']


class VispyLayerState(LayerState):
    """
    A base state object for all Vispy layers
    """

    color = CallbackProperty()
    alpha = CallbackProperty()

    def __init__(self, **kwargs):

        super(VispyLayerState, self).__init__(**kwargs)

        self._sync_color = None
        self._sync_alpha = None

        self.add_callback('layer', self._layer_changed)
        self._layer_changed()

        self.add_global_callback(self._notify_layer_update)

    def _notify_layer_update(self, **kwargs):
        message = LayerArtistUpdatedMessage(self)
        if self.layer is not None and self.layer.hub is not None:
            self.layer.hub.broadcast(message)

    def _layer_changed(self):

        if self._sync_color is not None:
            self._sync_color.stop_syncing()

        if self._sync_alpha is not None:
            self._sync_alpha.stop_syncing()

        if self.layer is not None:

            self.color = self.layer.style.color
            self.alpha = self.layer.style.alpha

            self._sync_color = keep_in_sync(self, 'color', self.layer.style, 'color')
            self._sync_alpha = keep_in_sync(self, 'alpha', self.layer.style, 'alpha')



# -*- coding: utf-8 -*-
# Copyright (c) 2015, Vispy Development Team.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.

""" This module provides an template for creating backends for vispy.
It clearly indicates what methods should be implemented and what events
should be emitted.
"""

from __future__ import division

from ..base import (BaseApplicationBackend, BaseCanvasBackend,
                    BaseTimerBackend)
from ...util import keys
from ... import config

USE_EGL = config['gl_backend'].lower().startswith('es')


# -------------------------------------------------------------------- init ---

# Map native keys to vispy keys
KEYMAP = {
    -1: keys.SHIFT,
    -2: keys.CONTROL,
    -3: keys.ALT,
    -4: keys.META,

    -5: keys.LEFT,
    -6: keys.UP,
    -7: keys.RIGHT,
    -8: keys.DOWN,
    -9: keys.PAGEUP,
    -10: keys.PAGEDOWN,

    -11: keys.INSERT,
    -12: keys.DELETE,
    -13: keys.HOME,
    -14: keys.END,

    -15: keys.ESCAPE,
    -16: keys.BACKSPACE,

    -17: keys.SPACE,
    -18: keys.ENTER,
    -19: keys.TAB,

    -20: keys.F1,
    -21: keys.F2,
    -22: keys.F3,
    -23: keys.F4,
    -24: keys.F5,
    -25: keys.F6,
    -26: keys.F7,
    -27: keys.F8,
    -28: keys.F9,
    -29: keys.F10,
    -30: keys.F11,
    -31: keys.F12,
}


# -------------------------------------------------------------- capability ---

# These are all booleans. Note that they mirror many of the kwargs to
# the initialization of the Canvas class.
capability = dict(
    # if True they mean:
    title=False,          # can set title on the fly
    size=False,           # can set size on the fly
    position=False,       # can set position on the fly
    show=False,           # can show/hide window XXX ?
    vsync=False,          # can set window to sync to blank
    resizable=False,      # can toggle resizability (e.g., no user resizing)
    decorate=False,       # can toggle decorations
    fullscreen=False,     # fullscreen window support
    context=False,        # can share contexts between windows
    multi_window=False,   # can use multiple windows at once
    scroll=False,         # scroll-wheel events are supported
    parent=False,         # can pass native widget backend parent
    always_on_top=False,  # can be made always-on-top
)


# ------------------------------------------------------- set_configuration ---
def _set_config(c):
    """Set gl configuration for template"""
    raise NotImplementedError


# ------------------------------------------------------------- application ---

class ApplicationBackend(BaseApplicationBackend):

    def __init__(self):
        BaseApplicationBackend.__init__(self)

    def _vispy_get_backend_name(self):
        return 'ThisBackendsName'

    def _vispy_process_events(self):
        raise NotImplementedError()

    def _vispy_run(self):
        raise NotImplementedError()

    def _vispy_quit(self):
        raise NotImplementedError()

    def _vispy_get_native_app(self):
        raise NotImplementedError()


# ------------------------------------------------------------------ canvas ---

# You can mix this class with the native widget
class CanvasBackend(BaseCanvasBackend):
    """Template backend

    Events to emit are shown below. Most backends will probably
    have one method for each event:

        self._vispy_canvas.events.initialize()
        self._vispy_canvas.events.resize(size=(w, h))
        self._vispy_canvas.events.draw(region=None)
        self._vispy_canvas.close()
        self._vispy_canvas.events.mouse_press(pos=(x, y), button=1,
                                              modifiers=())
        self._vispy_canvas.events.mouse_release(pos=(x, y), button=1,
                                                modifiers=())
        self._vispy_canvas.events.mouse_double_click(pos=(x, y), button=1,
                                                     modifiers=())
        self._vispy_canvas.events.mouse_move(pos=(x, y), modifiers=())
        self._vispy_canvas.events.mouse_wheel(pos=(x, y), delta=(0, 0),
                                              modifiers=())
        self._vispy_canvas.events.key_press(key=key, text=text, modifiers=())
        self._vispy_canvas.events.key_release(key=key, text=text, modifiers=())

    In most cases, if the window-cross is clicked, a native close-event is
    generated, which should then call canvas.close(). The Canvas class is
    responsible for firing the close event and calling
    backend_canvas._vispy_close, which closes the native widget.
    If this happens to result in a second close event, canvas.close() gets
    called again, but Canvas knows it is closing so it stops there.

    If canvas.close() is called (by the user), it calls
    backend_canvas._vispy_close, which closes the native widget,
    and we get the same stream of actions as above. This deviation from
    having events come from the CanvasBackend is necessitated by how
    different backends handle close events, and the various ways such
    events can be triggered.
    """

    # args are for BaseCanvasBackend, kwargs are for us.
    def __init__(self, *args, **kwargs):
        BaseCanvasBackend.__init__(self, *args)
        # We use _process_backend_kwargs() to "serialize" the kwargs
        # and to check whether they match this backend's capability
        p = self._process_backend_kwargs(kwargs)

        # Deal with config
        # ... use context.config
        # Deal with context
        p.context.shared.add_ref('backend-name', self)
        if p.context.shared.ref is self:
            self._native_context = None  # ...
        else:
            self._native_context = p.context.shared.ref._native_context

        # NativeWidgetClass.__init__(self, foo, bar)

    def _vispy_set_current(self):
        # Make this the current context
        raise NotImplementedError()

    def _vispy_swap_buffers(self):
        # Swap front and back buffer
        raise NotImplementedError()

    def _vispy_set_title(self, title):
        # Set the window title. Has no effect for widgets
        raise NotImplementedError()

    def _vispy_set_size(self, w, h):
        # Set size of the widget or window
        raise NotImplementedError()

    def _vispy_set_position(self, x, y):
        # Set location of the widget or window. May have no effect for widgets
        raise NotImplementedError()

    def _vispy_set_visible(self, visible):
        # Show or hide the window or widget
        raise NotImplementedError()

    def _vispy_set_fullscreen(self, fullscreen):
        # Set the current fullscreen state
        raise NotImplementedError()

    def _vispy_update(self):
        # Invoke a redraw
        raise NotImplementedError()

    def _vispy_close(self):
        # Force the window or widget to shut down
        raise NotImplementedError()

    def _vispy_get_size(self):
        # Should return widget size
        raise NotImplementedError()

    def _vispy_get_position(self):
        # Should return widget position
        raise NotImplementedError()

    def _vispy_get_fullscreen(self):
        # Should return the current fullscreen state
        raise NotImplementedError()

    def _vispy_get_native_canvas(self):
        # Should return the native widget object.
        # If this is self, this method can be omitted.
        return self


# ------------------------------------------------------------------- timer ---

class TimerBackend(BaseTimerBackend):  # Can be mixed with native timer class

    def __init__(self, vispy_timer):
        BaseTimerBackend.__init__(self, vispy_timer)

    def _vispy_start(self, interval):
        raise NotImplementedError()

    def _vispy_stop(self):
        raise NotImplementedError()

    def _vispy_timeout(self):
        raise NotImplementedError()

    def _vispy_get_native_timer(self):
        # Should return the native widget object.
        # If this is self, this method can be omitted.
        return self



# This file implements a MultiVolumeVisual class that can be used to show
# multiple volumes simultaneously. It is derived from the original VolumeVisual
# class in vispy.visuals.volume, which is releaed under a BSD license included
# here:
#
# ===========================================================================
# Vispy is licensed under the terms of the (new) BSD license:
#
# Copyright (c) 2015, authors of Vispy
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# * Redistributions of source code must retain the above copyright
#   notice, this list of conditions and the following disclaimer.
# * Redistributions in binary form must reproduce the above copyright
#   notice, this list of conditions and the following disclaimer in the
#   documentation and/or other materials provided with the distribution.
# * Neither the name of Vispy Development Team nor the names of its
#   contributors may be used to endorse or promote products
#   derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
# IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
# TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER
# OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
# ===========================================================================
#
# This modified version is released under the BSD license given in the LICENSE
# file in this repository.

from __future__ import absolute_import, division, print_function

from distutils.version import LooseVersion
from collections import defaultdict

import numpy as np
from glue.external import six
from glue.utils import iterate_chunks

from ..extern.vispy.gloo import Texture3D, TextureEmulated3D, VertexBuffer, IndexBuffer
from ..extern.vispy.visuals import VolumeVisual, Visual
from ..extern.vispy.visuals.shaders import Function
from ..extern.vispy.color import get_colormap, Color
from ..extern.vispy.scene.visuals import create_visual_node

from .shaders import get_frag_shader, VERT_SHADER

NUMPY_LT_1_13 = LooseVersion(np.__version__) < LooseVersion('1.13')


class NoFreeSlotsError(Exception):
    pass


class MultiVolumeVisual(VolumeVisual):
    """
    Displays multiple 3D volumes simultaneously.

    Parameters
    ----------
    volumes : list of tuples
        The volumes to show. Each tuple should contain three elements: the data
        array, the clim values, and the colormap to use. The clim values should
        be either a 2-element tuple, or None.
    emulate_texture : bool
        Use 2D textures to emulate a 3D texture. OpenGL ES 2.0 compatible,
        but has lower performance on desktop platforms.
    n_volume_max : int
        Absolute maximum number of volumes that can be shown.
    """

    def __init__(self, n_volume_max=16, emulate_texture=False, bgcolor='white', resolution=256):

        # Choose texture class
        tex_cls = TextureEmulated3D if emulate_texture else Texture3D

        self._n_volume_max = n_volume_max
        self._vol_shape = (resolution, resolution, resolution)
        self._need_vertex_update = True
        self._data_bounds = None

        self.resolution = resolution

        # We deliberately don't use super here because we don't want to call
        # VolumeVisual.__init__
        Visual.__init__(self, vcode=VERT_SHADER, fcode="")

        self.volumes = defaultdict(dict)

        # We turn on clipping straight away - the following variable is needed
        # by _update_shader
        self._clip_data = True

        # Set up initial shader so that we can start setting shader variables
        # that don't depend on what volumes are actually active.
        self._update_shader()

        # Set initial clipping parameters
        self.shared_program['u_clip_min'] = [0, 0, 0]
        self.shared_program['u_clip_max'] = [1, 1, 1]

        # Set up texture vertices - note that these variables are required by
        # the parent VolumeVisual class.

        self._vertices = VertexBuffer()
        self._texcoord = VertexBuffer(
            np.array([[0, 0, 0],
                      [1, 0, 0],
                      [0, 1, 0],
                      [1, 1, 0],
                      [0, 0, 1],
                      [1, 0, 1],
                      [0, 1, 1],
                      [1, 1, 1]], dtype=np.float32))

        self._draw_mode = 'triangle_strip'
        self._index_buffer = IndexBuffer()

        self.shared_program['a_position'] = self._vertices
        self.shared_program['a_texcoord'] = self._texcoord

        # Only show back faces of cuboid. This is required because if we are
        # inside the volume, then the front faces are outside of the clipping
        # box and will not be drawn.
        self.set_gl_state('translucent', cull_face=False)

        # Set up the underlying volume shape and define textures

        self._vol_shape = (resolution, resolution, resolution)
        self.shared_program['u_shape'] = self._vol_shape

        self.textures = []
        for i in range(n_volume_max):

            # Set up texture object
            self.textures.append(tex_cls(self._vol_shape, interpolation='linear',
                                         wrapping='clamp_to_edge'))

            # Pass texture object to shader program
            self.shared_program['u_volumetex_{0}'.format(i)] = self.textures[i]

            # Make sure all textures are disabled
            self.shared_program['u_enabled_{0}'.format(i)] = 0
            self.shared_program['u_weight_{0}'.format(i)] = 1

        # Don't use downsampling initially (1 means show 1:1 resolution)
        self.shared_program['u_downsample'] = 1.

        # Set up texture sampler
        self.shared_program.frag['sampler_type'] = self.textures[0].glsl_sampler_type
        self.shared_program.frag['sample'] = self.textures[0].glsl_sample

        # Set initial background color
        self.shared_program['u_bgcolor'] = Color(bgcolor).rgba

        # Prevent additional attributes from being added
        try:
            self.freeze()
        except AttributeError:  # Older versions of VisPy
            pass

    def _update_shader(self, force=False):
        shader = get_frag_shader(self.volumes, clipped=self._clip_data,
                                 n_volume_max=self._n_volume_max)
        # We only actually update the shader in OpenGL if the code has changed
        # to avoid any overheads in uploading the new shader code
        if force or getattr(self, '_shader_cache', None) != shader:
            self.shared_program.frag = shader
            self._shader_cache = shader

    # The following methods change things which require the shader code to be updated

    def allocate(self, label):
        if label in self.volumes:
            raise ValueError("Label {0} already exists".format(label))
        index = self._free_slot_index
        self.volumes[label] = {}
        self.volumes[label]['index'] = index
        self.shared_program['u_enabled_{0}'.format(index)] = 0
        self._update_shader()

    def deallocate(self, label):
        if label not in self.volumes:
            return  # layer already deallocated
        self.disable(label)
        self.volumes.pop(label)
        self._update_shader()

    def set_clip(self, clip_data, clip_limits):
        self._clip_data = int(clip_data)
        if clip_data:
            self.shared_program['u_clip_min'] = clip_limits[:3]
            self.shared_program['u_clip_max'] = clip_limits[3:]
        self._update_shader()

    def set_multiply(self, label, label_other):
        self.volumes[label]['multiply'] = label_other
        self._update_shader()

    # The following methods don't require any changes to the shader code, so we
    # don't update the shader after setting the OpenGL variables.

    def enable(self, label):
        index = self.volumes[label]['index']
        self.shared_program['u_enabled_{0}'.format(index)] = 1

    def disable(self, label):
        index = self.volumes[label]['index']
        self.shared_program['u_enabled_{0}'.format(index)] = 0

    def downsample(self):
        min_dimension = min(self._vol_shape)
        self.shared_program['u_downsample'] = min_dimension / 20

    def upsample(self):
        self.shared_program['u_downsample'] = 1.

    def set_background(self, color):
        self.shared_program['u_bgcolor'] = Color(color).rgba

    def set_resolution(self, resolution):
        self.resolution = resolution
        self._vol_shape = (resolution, resolution, resolution)
        self.shared_program['u_shape'] = self._vol_shape[::-1]

    def set_cmap(self, label, cmap):
        if isinstance(cmap, six.string_types):
            cmap = get_colormap(cmap)
        self.volumes[label]['cmap'] = cmap
        index = self.volumes[label]['index']
        self.shared_program.frag['cmap{0:d}'.format(index)] = Function(cmap.glsl_map)

    def set_clim(self, label, clim):
        # Avoid setting the same limits again
        if 'clim' in self.volumes[label] and self.volumes[label]['clim'] == clim:
            return
        self.volumes[label]['clim'] = clim
        if 'data' in self.volumes[label]:
            self._update_scaled_data(label)

    def set_weight(self, label, weight):
        index = self.volumes[label]['index']
        self.shared_program['u_weight_{0:d}'.format(index)] = weight

    def set_data(self, label, data, layer=None):

        if 'clim' not in self.volumes[label]:
            raise ValueError("set_clim should be called before set_data")

        # Avoid adding the same data again
        if 'data' in self.volumes[label] and self.volumes[label]['data'] is data:
            return

        self.volumes[label]['data'] = data
        self.volumes[label]['layer'] = layer
        self._update_scaled_data(label)

    def _update_scaled_data(self, label):

        # If the data slice hasn't been set yet, we should stop here
        if self._data_bounds is None:
            return

        index = self.volumes[label]['index']
        clim = self.volumes[label].get('clim', None)
        data = self.volumes[label]['data']

        # With certain graphics cards, sending the data in one chunk to OpenGL
        # causes artifacts in the rendering - see e.g.
        # https://github.com/vispy/vispy/issues/1412
        # To avoid this, we process the data in chunks. Since we need to do
        # this, we can also do the copy and renormalization on the chunk to
        # avoid excessive memory usage.

        # To start off we need to tell the texture about the new shape
        self.shared_program['u_volumetex_{0:d}'.format(index)].resize(data.shape)

        # Determine the chunk shape - the value of 128 as the minimum value
        # is arbitrary but appears to work nicely. We can reduce that in future
        # if needed.

        sliced_data = data.compute_fixed_resolution_buffer(self._data_bounds)

        chunk_shape = [min(x, 128, self.resolution) for x in sliced_data.shape]

        # FIXME: shouldn't be needed!
        zeros = np.zeros(self._vol_shape, dtype=np.float32)
        self.shared_program['u_volumetex_{0:d}'.format(index)].set_data(zeros)

        # Now loop over chunks

        for view in iterate_chunks(sliced_data.shape, chunk_shape=chunk_shape):

            chunk = sliced_data[view]
            chunk = chunk.astype(np.float32)
            if clim is not None:
                chunk -= clim[0]
                chunk *= 1 / (clim[1] - clim[0])

            # PERF: nan_to_num doesn't actually help memory usage as it runs
            # isnan internally, and it's slower, so we just use the following
            # methind. In future we could do this directly with a C extension.
            chunk[np.isnan(chunk)] = 0.

            offset = tuple([s.start for s in view])

            if chunk.size == 0:
                continue

            self.shared_program['u_volumetex_{0:d}'.format(index)].set_data(chunk, offset=offset)

    def label_for_layer(self, layer):
        for label in self.volumes:
            if 'layer' in self.volumes[label]:
                if self.volumes[label]['layer'] is layer:
                    return label

    @property
    def has_free_slots(self):
        try:
            self._free_slot_index
        except NoFreeSlotsError:
            return False
        else:
            return True

    @property
    def _free_slot_index(self):
        indices = [self.volumes[label]['index'] for label in self.volumes]
        for index in range(self._n_volume_max):
            if index not in indices:
                return index
        raise NoFreeSlotsError("No free slots")

    def _update_slice_transform(self, x_min, x_max, y_min, y_max, z_min, z_max):

        # TODO: simplify this to get bounds, for FRB

        x_step = (x_max - x_min) / self.resolution
        y_step = (y_max - y_min) / self.resolution
        z_step = (z_max - z_min) / self.resolution

        data_bounds = [(z_min, z_max, self.resolution),
                       (y_min, y_max, self.resolution),
                       (x_min, x_max, self.resolution)]

        # We should stop at this point if the bounds are the same as before
        if data_bounds == self._data_bounds:
            return
        else:
            self._data_bounds = data_bounds

        self.transform.inner.scale = [x_step, y_step, z_step]
        self.transform.inner.translate = [x_min, y_min, z_min]

        # We need to update the data in OpenGL if the slice has changed
        for label in self.volumes:
            self._update_scaled_data(label)

        # The following is needed to make sure that VisPy recognizes the changes
        # to the transforms.
        self.transform._update_shaders()
        self.transform.update()

    @property
    def enabled(self):
        return [self.shared_program['u_enabled_{0}'.format(i)] == 1
                for i in range(self._n_volume_max)]

    def draw(self):
        if not any(self.enabled):
            return
        else:
            try:
                super(MultiVolumeVisual, self).draw()
            except Exception:
                pass


MultiVolume = create_visual_node(MultiVolumeVisual)



# Make sure that session files can be read in a backward-compatible manner

from __future__ import absolute_import, division, print_function

import os
import pytest
from mock import patch

import numpy as np
from numpy.testing import assert_equal

from glue.core.state import GlueUnSerializer

from ..volume.volume_viewer import QMessageBox

DATA = os.path.join(os.path.dirname(__file__), 'data')


@pytest.mark.parametrize('protocol', [0, 1])
def test_scatter_volume(protocol):

    filename = os.path.join(DATA, 'scatter_volume_v{0}.glu'.format(protocol))

    with open(filename, 'r') as f:
        session = f.read()

    with patch.object(QMessageBox, 'question') as question:
        question.return_value = QMessageBox.Yes
        state = GlueUnSerializer.loads(session)
        ga = state.object('__main__')

    dc = ga.session.data_collection

    assert len(dc) == 2

    assert dc[0].label == 'table'
    assert dc[1].label == 'array'

    # SCATTER VIEWER

    scatter = ga.viewers[0][0]

    viewer_state = scatter.state

    assert viewer_state.x_att.label == 'b'
    assert viewer_state.x_min == 8
    assert viewer_state.x_max == 0
    np.testing.assert_allclose(viewer_state.x_stretch, 0.4, rtol=1.e-3)

    assert viewer_state.y_att.label == 'a'
    assert viewer_state.y_min == 0.4
    assert viewer_state.y_max == 4
    np.testing.assert_allclose(viewer_state.y_stretch, 0.6, rtol=1.e-3)

    assert viewer_state.z_att.label == 'c'
    assert viewer_state.z_min == 2
    assert viewer_state.z_max == 5.4
    np.testing.assert_allclose(viewer_state.z_stretch, 1.4, rtol=1.e-3)

    assert viewer_state.perspective_view
    assert not viewer_state.visible_axes
    if protocol >= 1:
        assert not viewer_state.native_aspect
        assert viewer_state.clip_data

    layer_state = viewer_state.layers[0]

    assert layer_state.size_mode == 'Linear'
    assert layer_state.size_attribute.label == 'e'
    assert layer_state.size_vmin == 3
    assert layer_state.size_vmax == 4
    np.testing.assert_allclose(layer_state.size_scaling, 1.51356, rtol=1.e-3)

    assert layer_state.color_mode == 'Linear'
    assert layer_state.cmap_attribute.label == 'd'
    assert layer_state.cmap_vmin == 0
    assert layer_state.cmap_vmax == 10
    assert layer_state.alpha == 0.66

    # VOLUME VIEWER

    volume = ga.viewers[0][1]

    viewer_state = volume.state

    assert viewer_state.x_att.label == 'Pixel Axis 2 [x]'
    assert viewer_state.x_min == 3.5
    assert viewer_state.x_max == -0.5
    np.testing.assert_allclose(viewer_state.x_stretch, 0.2, rtol=1.e-3)

    assert viewer_state.y_att.label == 'Pixel Axis 1 [y]'
    assert viewer_state.y_min == -0.5
    assert viewer_state.y_max == 2.5
    np.testing.assert_allclose(viewer_state.y_stretch, 1.2, rtol=1.e-3)

    assert viewer_state.z_att.label == 'Pixel Axis 0 [z]'
    assert viewer_state.z_min == -0.5
    assert viewer_state.z_max == 1.5
    np.testing.assert_allclose(viewer_state.z_stretch, 1.3, rtol=1.e-3)

    assert not viewer_state.perspective_view
    assert viewer_state.visible_axes
    if protocol >= 1:
        assert viewer_state.native_aspect
        assert not viewer_state.clip_data

    layer_state = viewer_state.layers[0]

    assert layer_state.attribute == 'array'
    assert layer_state.vmin == 0
    assert layer_state.vmax == 23
    assert layer_state.color == '#e60010'
    assert layer_state.alpha == 0.36

    ga.close()


def test_scatter_volume_selection():

    filename = os.path.join(DATA, 'scatter_volume_selection.glu')

    with open(filename, 'r') as f:
        session = f.read()

    with patch.object(QMessageBox, 'question') as question:
        question.return_value = QMessageBox.Yes
        state = GlueUnSerializer.loads(session)
        ga = state.object('__main__')

    dc = ga.session.data_collection

    assert len(dc) == 2

    assert dc[0].label == 'array'
    assert dc[1].label == 'table'

    expected_array = np.array([[[0, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 0, 0, 0]],
                               [[0, 0, 0, 0],
                                [0, 0, 0, 0],
                                [0, 1, 0, 0],
                                [0, 1, 1, 1]],
                               [[0, 0, 0, 0],
                                [0, 0, 1, 1],
                                [0, 1, 1, 1],
                                [0, 1, 1, 1]],
                               [[0, 0, 1, 1],
                                [0, 0, 1, 1],
                                [0, 1, 1, 1],
                                [0, 0, 0, 1]]], dtype=bool)

    expected_table = np.array([0, 1, 0], dtype=bool)

    assert_equal(dc[0].subsets[0].to_mask(), expected_array)
    assert_equal(dc[1].subsets[0].to_mask(), expected_table)

    ga.close()


@pytest.mark.parametrize('protocol', [1, 2])
def test_multiple_volumes(protocol):

    # Before glue-vispy-viewers 0.12, volumes could be shown together without
    # being linked, so when loading old session files we should suggest to
    # auto link.

    filename = os.path.join(DATA, 'multiple_volumes_v{0}.glu'.format(protocol))

    with open(filename, 'r') as f:
        session = f.read()

    with patch.object(QMessageBox, 'question') as question:
        question.return_value = QMessageBox.Yes
        state = GlueUnSerializer.loads(session)
        ga = state.object('__main__')

    volume = ga.viewers[0][0]

    assert volume.layers[0].enabled
    assert volume.layers[1].enabled
    assert volume.layers[2].enabled
    assert volume.layers[3].enabled
    assert volume.layers[4].enabled
    assert volume.layers[5].enabled

    ga.close()



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014, Nicolas P. Rougier. All rights reserved.
# Distributed under the terms of the new BSD License.
# -----------------------------------------------------------------------------

from . length import XLength, YLength


class Viewport(object):

    def __init__(self, content=None, parent=None):

        self._x = None
        self._computed_x = 0
        if content.get('x'):
            self._x = XLength(content.get('x'), parent)
            self._computed_x = float(self._x)

        self._y = None
        self._computed_y = 0
        if content.get('y'):
            self._y = XLength(content.get('y'), parent)
            self._computed_y = float(self._y)

        self._width = None
        self._computed_width = 800
        if content.get('width'):
            self._width = XLength(content.get('width'), parent)
            self._computed_width = float(self._width)

        self._height = None
        self._computed_height = 800
        if content.get('height'):
            self._height = YLength(content.get('height'), parent)
            self._computed_height = float(self._height)

    @property
    def x(self):
        return self._computed_x

    @property
    def y(self):
        return self._computed_y

    @property
    def width(self):
        return self._computed_width

    @property
    def height(self):
        return self._computed_height

    def __repr__(self):
        s = repr((self._x, self._y, self._width, self._height))
        return s

    @property
    def xml(self):
        return self._xml

    @property
    def _xml(self, prefix=""):
        s = ""
        if self._x:
            s += 'x="%s" ' % repr(self._x)
        if self._y:
            s += 'y="%s" ' % repr(self._y)
        if self._width:
            s += 'width="%s" ' % repr(self._width)
        if self._height:
            s += 'height="%s" ' % repr(self._height)
        return s



from __future__ import absolute_import, division, print_function

import sys

import numpy as np

from ..extern.vispy import scene
from .axes import AxesVisual3D
from ..utils import NestedSTTransform

from matplotlib.colors import ColorConverter

from glue.config import settings

rgb = ColorConverter().to_rgb

LIMITS_PROPS = [coord + attribute for coord in 'xyz' for attribute in ['_min', '_max', '_stretch']]


class VispyWidgetHelper(object):

    def __init__(self, parent=None, viewer_state=None):

        # Prepare Vispy canvas. We set the depth_size to 24 to avoid issues
        # with isosurfaces on MacOS X
        self.canvas = scene.SceneCanvas(keys=None, show=False,
                                        config={'depth_size': 24},
                                        bgcolor=rgb(settings.BACKGROUND_COLOR))

        # Set up a viewbox
        self.view = self.canvas.central_widget.add_view()
        self.view.parent = self.canvas.scene

        # Set whether we are emulating a 3D texture. This needs to be enabled
        # as a workaround on Windows otherwise VisPy crashes.
        self.emulate_texture = (sys.platform == 'win32' and
                                sys.version_info[0] < 3)

        self.scene_transform = scene.STTransform()
        self.limit_transforms = {}

        fc = rgb(settings.FOREGROUND_COLOR)

        self.axis = AxesVisual3D(axis_color=fc, tick_color=fc, text_color=fc,
                                 tick_width=1, minor_tick_length=2,
                                 major_tick_length=4, axis_width=0,
                                 tick_label_margin=10, axis_label_margin=25,
                                 tick_font_size=6, axis_font_size=8,
                                 view=self.view,
                                 transform=self.scene_transform)

        # Create a turntable camera. For now, this is the only camerate type
        # we support, but if we support more in future, we should implement
        # that here

        # Orthographic perspective view as default
        self.view.camera = scene.cameras.TurntableCamera(parent=self.view.scene,
                                                         fov=0., distance=4.0)

        # We need to call render here otherwise we'll later encounter an OpenGL
        # program validation error.
        # self.canvas.render()

        self.viewer_state = viewer_state
        try:
            self.viewer_state.add_callback('*', self._update_from_state, as_kwargs=True)
        except TypeError:  # glue-core >= 0.11
            self.viewer_state.add_global_callback(self._update_from_state)
        self._update_from_state(force=True)

    def _update_appearance_from_settings(self):
        self.canvas.bgcolor = rgb(settings.BACKGROUND_COLOR)
        self.axis.axis_color = rgb(settings.FOREGROUND_COLOR)
        self.axis.tick_color = rgb(settings.FOREGROUND_COLOR)
        self.axis.label_color = rgb(settings.FOREGROUND_COLOR)

    def add_data_visual(self, visual):
        self.limit_transforms[visual] = NestedSTTransform()
        self._update_limits()
        visual.transform = self.limit_transforms[visual]
        self.view.add(visual)

    def _update_from_state(self, force=False, **props):

        if force or 'visible_axes' in props:
            self._toggle_axes()

        if force or 'perspective_view' in props:
            self._toggle_perspective()

        if force or any(key in props for key in ('x_att', 'y_att', 'z_att')):
            self._update_attributes()

        if force or any(key in props for key in ('x_stretch', 'y_stretch',
                                                 'z_stretch', 'native_aspect')):
            self._update_stretch()

        if force or any(p in props for p in LIMITS_PROPS) or 'native_aspect' in props:
            self._update_limits()

        self.canvas.update()

    def _toggle_axes(self):
        if self.viewer_state.visible_axes:
            self.axis.parent = self.view.scene
        else:
            self.axis.parent = None

    def _toggle_perspective(self):
        if self.viewer_state.perspective_view:
            self.view.camera.fov = 30
            self.axis.tick_font_size = 28
            self.axis.axis_font_size = 35
        else:
            self.view.camera.fov = 0
            self.axis.tick_font_size = 6
            self.axis.axis_font_size = 8

    def _update_attributes(self):
        if self.viewer_state.x_att is not None:
            self.axis.xlabel = self.viewer_state.x_att.label
        if self.viewer_state.y_att is not None:
            self.axis.ylabel = self.viewer_state.y_att.label
        if self.viewer_state.z_att is not None:
            self.axis.zlabel = self.viewer_state.z_att.label

    def _update_stretch(self):
        self.scene_transform.scale = (self.viewer_state.x_stretch * self.viewer_state.aspect[0],
                                      self.viewer_state.y_stretch * self.viewer_state.aspect[1],
                                      self.viewer_state.z_stretch * self.viewer_state.aspect[2])

    def _update_limits(self):

        dx = self.viewer_state.x_max - self.viewer_state.x_min
        sx = (np.inf if dx == 0 else 2. / dx *
              self.viewer_state.x_stretch * self.viewer_state.aspect[0])

        dy = self.viewer_state.y_max - self.viewer_state.y_min
        sy = (np.inf if dy == 0 else 2. / dy *
              self.viewer_state.y_stretch * self.viewer_state.aspect[1])

        dz = self.viewer_state.z_max - self.viewer_state.z_min
        sz = (np.inf if dz == 0 else 2. / dz *
              self.viewer_state.z_stretch * self.viewer_state.aspect[2])

        scale = [sx, sy, sz]

        translate = [-0.5 * (self.viewer_state.x_min + self.viewer_state.x_max) * scale[0],
                     -0.5 * (self.viewer_state.y_min + self.viewer_state.y_max) * scale[1],
                     -0.5 * (self.viewer_state.z_min + self.viewer_state.z_max) * scale[2]]

        for visual in self.limit_transforms:
            self.limit_transforms[visual].scale = scale
            self.limit_transforms[visual].translate = translate

        self.axis.xlim = self.viewer_state.x_min, self.viewer_state.x_max
        self.axis.ylim = self.viewer_state.y_min, self.viewer_state.y_max
        self.axis.zlim = self.viewer_state.z_min, self.viewer_state.z_max



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2015, Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------

from ..ext.six import string_types
from .wrappers import read_pixels


def _screenshot(viewport=None, alpha=True):
    """ Take a screenshot using glReadPixels. Not sure where to put this
    yet, so a private function for now. Used in make.py.

    Parameters
    ----------
    viewport : array-like | None
        4-element list of x, y, w, h parameters. If None (default),
        the current GL viewport will be queried and used.
    alpha : bool
        If True (default), the returned array has 4 elements (RGBA).
        Otherwise, it has 3 (RGB)

    Returns
    -------
    pixels : array
        3D array of pixels in np.uint8 format
    """
    # gl.glReadBuffer(gl.GL_BACK)  Not avaliable in ES 2.0
    return read_pixels(viewport, alpha)


KEYWORDS = set(['active', 'asm', 'cast', 'class', 'common', 'default',
                'double', 'dvec2', 'dvec3', 'dvec4', 'enum', 'extern',
                'external', 'filter', 'fixed', 'flat', 'fvec2', 'fvec3',
                'fvec4', 'goto', 'half', 'hvec2', 'hvec3', 'hvec4', 'iimage1D',
                'iimage1DArray', 'iimage2D', 'iimage2DArray', 'iimage3D',
                'iimageBuffer', 'iimageCube', 'image1D', 'image1DArray',
                'image1DArrayShadow', 'image1DShadow', 'image2D',
                'image2DArray', 'image2DArrayShadow', 'image2DShadow',
                'image3D', 'imageBuffer', 'imageCube', 'inline', 'input',
                'interface', 'long', 'namespace', 'noinline', 'output',
                'packed', 'partition', 'public', 'row_major', 'sampler1D',
                'sampler1DShadow', 'sampler2DRect', 'sampler2DRectShadow',
                'sampler2DShadow', 'sampler3D', 'sampler3DRect', 'short',
                'sizeof', 'static', 'superp', 'switch', 'template', 'this',
                'typedef', 'uimage1D', 'uimage1DArray', 'uimage2D',
                'uimage2DArray', 'uimage3D', 'uimageBuffer', 'uimageCube',
                'union', 'unsigned', 'using', 'volatile'])


def check_variable(name):
    """
    Return None if *name* is expected to be a valid variable name in any GLSL
    version. Otherwise, return a string describing the problem.
    """
    # Limit imposed by glGetActive* in pyopengl
    if len(name) > 31:
        return ("Variable names >31 characters may not function on some "
                "systems.")

    return check_identifier(name)


def check_identifier(name):
    if '__' in name:
        return "Identifiers may not contain double-underscores."
    
    if name[:3] == 'gl_' or name[:3] == 'GL_':
        return "Identifiers may not begin with gl_ or GL_."

    if name in KEYWORDS:
        return "Identifier is a reserved keyword."


def check_enum(enum, name=None, valid=None):
    """ Get lowercase string representation of enum.
    """
    name = name or 'enum'
    # Try to convert
    res = None
    if isinstance(enum, int):
        if hasattr(enum, 'name') and enum.name.startswith('GL_'):
            res = enum.name[3:].lower()
    elif isinstance(enum, string_types):
        res = enum.lower()
    # Check
    if res is None:
        raise ValueError('Could not determine string represenatation for'
                         'enum %r' % enum)
    elif valid and res not in valid:
        raise ValueError('Value of %s must be one of %r, not %r' % 
                         (name, valid, enum))
    return res


vert_draw = """
attribute vec2 a_position;
attribute vec2 a_texcoord;
varying vec2 v_uv;

void main(void) {
    v_uv = a_texcoord.xy;
    gl_Position = vec4(a_position, 0, 1);
}
"""

frag_draw = """
uniform sampler2D u_texture;
varying vec2 v_uv;

void main(void) {
    gl_FragColor = texture2D(u_texture, v_uv).rgba;
}
"""


def draw_texture(tex):
    """Draw a 2D texture to the current viewport

    Parameters
    ----------
    tex : instance of Texture2D
        The texture to draw.
    """
    from .program import Program
    program = Program(vert_draw, frag_draw)
    program['u_texture'] = tex
    program['a_position'] = [[-1., -1.], [-1., 1.], [1., -1.], [1., 1.]]
    program['a_texcoord'] = [[0., 1.], [0., 0.], [1., 1.], [1., 0.]]
    program.draw('triangle_strip')



# -*- coding: utf-8 -*-
# -----------------------------------------------------------------------------
# Copyright (c) 2014, Vispy Development Team. All Rights Reserved.
# Distributed under the (new) BSD License. See LICENSE.txt for more info.
# -----------------------------------------------------------------------------

from ..geometry import create_box
from .mesh import MeshVisual
from .visual import CompoundVisual


class BoxVisual(CompoundVisual):
    """Visual that displays a box.

    Parameters
    ----------
    width : float
        Box width.
    height : float
        Box height.
    depth : float
        Box depth.
    width_segments : int
        Box segments count along the width.
    height_segments : float
        Box segments count along the height.
    depth_segments : float
        Box segments count along the depth.
    planes: array_like
        Any combination of ``{'-x', '+x', '-y', '+y', '-z', '+z'}``
        Included planes in the box construction.
    vertex_colors : ndarray
        Same as for `MeshVisual` class. See `create_plane` for vertex ordering.
    face_colors : ndarray
        Same as for `MeshVisual` class. See `create_plane` for vertex ordering.
    color : Color
        The `Color` to use when drawing the cube faces.
    edge_color : tuple or Color
        The `Color` to use when drawing the cube edges. If `None`, then no
        cube edges are drawn.
    """

    def __init__(self, width=1, height=1, depth=1, width_segments=1,
                 height_segments=1, depth_segments=1, planes=None,
                 vertex_colors=None, face_colors=None,
                 color=(0.5, 0.5, 1, 1), edge_color=None, **kwargs):
        vertices, filled_indices, outline_indices = create_box(
            width, height, depth, width_segments, height_segments,
            depth_segments, planes)

        self._mesh = MeshVisual(vertices['position'], filled_indices,
                                vertex_colors, face_colors, color)
        if edge_color:
            self._border = MeshVisual(vertices['position'], outline_indices,
                                      color=edge_color, mode='lines')
        else:
            self._border = MeshVisual()

        CompoundVisual.__init__(self, [self._mesh, self._border], **kwargs)
        self.mesh.set_gl_state(polygon_offset_fill=True,
                               polygon_offset=(1, 1), depth_test=True)

    @property
    def mesh(self):
        """The vispy.visuals.MeshVisual that used to fill in.
        """
        return self._mesh

    @mesh.setter
    def mesh(self, mesh):
        self._mesh = mesh

    @property
    def border(self):
        """The vispy.visuals.MeshVisual that used to draw the border.
        """
        return self._border

    @border.setter
    def border(self, border):
        self._border = border
